import jax.numpy as jnp
import jax
from jax.typing import ArrayLike, DTypeLike
from jax import Array

"""Universal functions for two-body orbital mechanics."""


def sigma_bvp(pos_vec: ArrayLike, vel_vec: ArrayLike, mu: ArrayLike = 1) -> Array:
    r"""The sigma function

    Args:
        pos_vec: (...,3) The position vector.
        vel_vec: (...,3) The velocity vector.
        mu: The gravitational parameter.

    Returns:
        The value of the sigma function.

    Notes:
        The sigma function is defined as:
        $$
        \sigma = \frac{\boldsymbol{r} \cdot \boldsymbol{v}}{\sqrt{\mu}}
        $$
        where $\boldsymbol{r}$ is the position vector, $\boldsymbol{v}$ is the velocity vector, and $\mu$ is the gravitational parameter.

    References:
        Battin, 1999, pp.174.

    Examples:
        A simple example of calculating the sigma function with a position vector of [1, 0, 0], a velocity vector of [0, 1, 0], and a gravitational parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> pos_vec = jnp.array([1.0, 0.0, 0.0])
        >>> vel_vec = jnp.array([0.0, 1.0, 0.0])
        >>> mu = 1.0
        >>> adx.twobody.sigma_bvp(pos_vec, vel_vec, mu)
        Array([0.], dtype=float32)

        With broadcasting, you can calculate the sigma function for multiple position and velocity vectors:

        >>> pos_vec = jnp.array([[1.0, 0.0, 0.0], [2.0, 0.0, 0.0]])
        >>> vel_vec = jnp.array([[0.0, 1.0, 0.0], [0.0, 2.0, 0.0]])
        >>> mu = jnp.array([[1.0], [2.0]])
        >>> adx.twobody.sigma_bvp(pos_vec, vel_vec, mu)
        Array([[0.],
               [0.]], dtype=float32)
    """
    return jnp.sum(pos_vec * vel_vec, axis=-1, keepdims=True) / jnp.sqrt(mu)


def ufunc0(chi: ArrayLike, alpha: DTypeLike) -> Array:
    r"""The universal function U0

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U0.

    Notes:
        The universal function U0 is defined as:
        $$
        U_0(\chi, \alpha) = \begin{cases}
        1 & \alpha = 0 \\
        \cos(\sqrt{\alpha} \chi) & \alpha > 0 \\
        \cosh(\sqrt{-\alpha} \chi) & \alpha < 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ufunc0(chi, alpha)
        Array(0.5403..., dtype=float32, weak_type=True)

        With broadcasting:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = 1.0
        >>> adx.twobody.ufunc0(chi, alpha)
        Array([ 0.5403..., -0.4161...], dtype=float32)
    """
    return jax.lax.cond(
        alpha > 0,
        lambda: jnp.cos(jnp.sqrt(alpha) * chi),
        lambda: jax.lax.cond(
            alpha < 0,
            lambda: jnp.cosh(jnp.sqrt(-alpha) * chi),
            lambda: jnp.ones_like(chi),
        ),
    )


def ufunc1(chi: ArrayLike, alpha: DTypeLike) -> Array:
    r"""The universal function U1

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U1.

    Notes:
        The universal function U1 is defined as:
        $$
        U_1(\chi, \alpha) = \begin{cases}
        \frac{\sin(\sqrt{\alpha} \chi)}{\sqrt{\alpha}} & \alpha > 0 \\
        \frac{\sinh(\sqrt{-\alpha} \chi)}{\sqrt{-\alpha}} & \alpha < 0 \\
        \chi & \alpha = 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ufunc1(chi, alpha)
        Array(0.8414..., dtype=float32, weak_type=True)

        With broadcasting:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = 1.0
        >>> adx.twobody.ufunc1(chi, alpha)
        Array([0.8414..., 0.9092...], dtype=float32)
    """
    return jax.lax.cond(
        alpha > 0,
        lambda: jnp.sin(jnp.sqrt(alpha) * chi) / jnp.sqrt(alpha),
        lambda: jax.lax.cond(
            alpha < 0,
            lambda: jnp.sinh(jnp.sqrt(-alpha) * chi) / jnp.sqrt(-alpha),
            lambda: chi,
        ),
    )


def ufunc2(chi: ArrayLike, alpha: DTypeLike) -> Array:
    r"""The universal function U2

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U2.

    Notes:
        The universal function U2 is defined as:
        $$
        U_2(\chi, \alpha) = \begin{cases}
        \frac{1 - \cos(\sqrt{\alpha} \chi)}{\alpha} & \alpha > 0 \\
        \frac{\cosh(\sqrt{-\alpha} \chi) - 1}{-\alpha} & \alpha < 0 \\
        \frac{\chi^2}{2} & \alpha = 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ufunc2(chi, alpha)
        Array(0.4596..., dtype=float32, weak_type=True)

        With broadcasting:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = 1.0
        >>> adx.twobody.ufunc2(chi, alpha)
        Array([0.4596..., 1.4161...], dtype=float32)
    """
    return jax.lax.cond(
        alpha > 0,
        lambda: (1 - jnp.cos(jnp.sqrt(alpha) * chi)) / alpha,
        lambda: jax.lax.cond(
            alpha < 0,
            lambda: (jnp.cosh(jnp.sqrt(-alpha) * chi) - 1) / -alpha,
            lambda: chi**2 / 2,
        ),
    )


def ufunc3(chi: ArrayLike, alpha: DTypeLike) -> Array:
    r"""The universal function U3

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U3.

    Notes:
        The universal function U3 is defined as:
        $$
        U_3(\chi, \alpha) = \begin{cases}
        \frac{\sqrt{\alpha} \chi - \sin(\sqrt{\alpha} \chi)}{\alpha \sqrt{\alpha}} & \alpha > 0 \\
        \frac{\sqrt{-\alpha} \chi - \sinh(\sqrt{-\alpha} \chi)}{\alpha \sqrt{-\alpha}} & \alpha < 0 \\
        \frac{\chi^3}{6} & \alpha = 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ufunc3(chi, alpha)
        Array(0.1585..., dtype=float32, weak_type=True)

        With broadcasting:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = 1.0
        >>> adx.twobody.ufunc3(chi, alpha)
        Array([0.1585..., 1.0907...], dtype=float32)
    """
    return jax.lax.cond(
        alpha > 0,
        lambda: (jnp.sqrt(alpha) * chi - jnp.sin(jnp.sqrt(alpha) * chi))
        / alpha
        / jnp.sqrt(alpha),
        lambda: jax.lax.cond(
            alpha < 0,
            lambda: (jnp.sqrt(-alpha) * chi - jnp.sinh(jnp.sqrt(-alpha) * chi))
            / alpha
            / jnp.sqrt(-alpha),
            lambda: chi**3 / 6,
        ),
    )


def ufunc4(chi: ArrayLike, alpha: DTypeLike) -> Array:
    r"""The universal function U4

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U4.

    Notes:
        The universal function U4 is defined as:
        $$
        U_4(\chi, \alpha) = \begin{cases}
        \frac{\alpha \chi^2 - 2 + 2 \cos(\sqrt{\alpha} \chi)}{2 \alpha^2} & \alpha > 0 \\
        \frac{\alpha \chi^2 - 2 + 2 \cosh(\sqrt{-\alpha} \chi)}{2 \alpha^2} & \alpha < 0 \\
        \frac{\chi^4}{24} & \alpha = 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.183.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ufunc4(chi, alpha)
        Array(0.0403..., dtype=float32, weak_type=True)

        With broadcasting:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = 1.0
        >>> adx.twobody.ufunc4(chi, alpha)
        Array([0.0403..., 0.5838...], dtype=float32)
    """
    return jax.lax.cond(
        alpha > 0,
        lambda: (alpha * chi**2 - 2 + 2 * jnp.cos(jnp.sqrt(alpha) * chi))
        / (2 * alpha**2),
        lambda: jax.lax.cond(
            alpha < 0,
            lambda: (alpha * chi**2 - 2 + 2 * jnp.cosh(jnp.sqrt(-alpha) * chi))
            / (2 * alpha**2),
            lambda: chi**4 / 24,
        ),
    )


def ufunc5(chi: ArrayLike, alpha: DTypeLike) -> Array:
    r"""The universal function U5

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U5.

    Notes:
        The universal function U5 is defined as:
        $$
        U_5(\chi, \alpha) = \begin{cases}
        \frac{\alpha^2 \chi^3 - 6\alpha \chi + 6\sqrt{\alpha} \sin(\sqrt{\alpha} \chi)}{6\alpha^3} & \alpha > 0 \\
        \frac{\alpha^2 \chi^3 - 6\alpha \chi - 6\sqrt{-\alpha} \sinh(\sqrt{-\alpha} \chi)}{6\alpha^3} & \alpha < 0 \\
        \frac{\chi^5}{120} & \alpha = 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.183.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ufunc5(chi, alpha)
        Array(0.0081..., dtype=float32, weak_type=True)

        With broadcasting:

        >>> chi = jnp.array([2.0, 3.0])
        >>> alpha = 1.0
        >>> adx.twobody.ufunc5(chi, alpha)
        Array([0.2426..., 1.6411...], dtype=float32)
    """
    return jax.lax.cond(
        alpha > 0,
        lambda: (
            alpha**2 * chi**3
            - 6 * alpha * chi
            + 6 * jnp.sqrt(alpha) * jnp.sin(jnp.sqrt(alpha) * chi)
        )
        / (6 * alpha**3),
        lambda: jax.lax.cond(
            alpha < 0,
            lambda: (
                alpha**2 * chi**3
                - 6 * alpha * chi
                - 6 * jnp.sqrt(-alpha) * jnp.sinh(jnp.sqrt(-alpha) * chi)
            )
            / (6 * alpha**3),
            lambda: chi**5 / 120,
        ),
    )



import jax.numpy as jnp
from jax.typing import ArrayLike
from jax import Array
from jax.numpy.linalg import vector_norm
import astrodynx as adx
from astrodynx.twobody._uniformulas import ufunc1, ufunc2, ufunc4, ufunc5, sigma_bvp
from astrodynx.twobody._lagrange import lagrange_F, lagrange_G, lagrange_Ft, lagrange_Gt

"""State transition matrices for two-body orbital mechanics."""


def _C_func(
    chi: ArrayLike,
    U2: ArrayLike,
    U4: ArrayLike,
    U5: ArrayLike,
    deltat: ArrayLike = 3.14,
    mu: ArrayLike = 1,
) -> ArrayLike:
    return (3 * U5 - chi * U4 - jnp.sqrt(mu) * deltat * U2) / jnp.sqrt(mu)


def prpr0(
    r_vec: ArrayLike,
    v_vec: ArrayLike,
    r0_vec: ArrayLike,
    v0_vec: ArrayLike,
    F: ArrayLike,
    C: ArrayLike,
    mu: ArrayLike = 1,
) -> Array:
    r"""The State Transition Matrix for position with respect to initial position.

    Args:
        r_vec: (3,) The position vector at the current time.
        v_vec: (3,) The velocity vector at the current time.
        r0_vec: (3,) The position vector at the initial time.
        v0_vec: (3,) The velocity vector at the initial time.
        F: (3,) The Lagrange F function.
        C: (3,) The C function.
        mu: (optional) The gravitational parameter.

    Returns:
        The State Transition Matrix for position with respect to initial position.

    Notes:
        The State Transition Matrix for position with respect to initial position is defined as:
        $$
        \frac{d\boldsymbol{r}}{d\boldsymbol{r}_0} =
        \frac{r}{\mu} (\boldsymbol{v} - \boldsymbol{v}_0) (\boldsymbol{v} - \boldsymbol{v}_0)^T
        + (1 - F) \frac{\boldsymbol{r} \boldsymbol{r}_0^T}{r_0^2}
        + \frac{C \boldsymbol{v} \boldsymbol{r}_0^T}{r_0^3}
        + F \boldsymbol{I}
        $$
        where $\boldsymbol{r}$ is the position vector at current time, $\boldsymbol{v}$ is the velocity vector at the current time, $\boldsymbol{r}_0$ is the position vector at the initial time, $\boldsymbol{v}_0$ is the velocity vector at the initial time, $F$ is the Lagrange F function, $C$ is the C function, $\mu$ is the gravitational parameter, and $\boldsymbol{I}$ is the identity matrix.

    References:
        Battin, 1999, pp.467.
    """
    r0_mag = vector_norm(r0_vec)
    return (
        vector_norm(r_vec) / mu * jnp.outer(v_vec - v0_vec, v_vec - v0_vec)
        + (1 - F) * jnp.outer(r_vec / r0_mag, r0_vec / r0_mag)
        + C / r0_mag**2 * jnp.outer(v_vec, r0_vec / r0_mag)
        + F * jnp.eye(3)
    )


def prpv0(
    r_vec: ArrayLike,
    v_vec: ArrayLike,
    r0_vec: ArrayLike,
    v0_vec: ArrayLike,
    F: ArrayLike,
    G: ArrayLike,
    C: ArrayLike,
    mu: ArrayLike = 1,
) -> Array:
    r"""The State Transition Matrix for position with respect to initial velocity.

    Args:
        r_vec: (3,) The position vector at the current time.
        v_vec: (3,) The velocity vector at the current time.
        r0_vec: (3,) The position vector at the initial time.
        v0_vec: (3,) The velocity vector at the initial time.
        F: (3,) The Lagrange F function.
        G: (3,) The Lagrange G function.
        C: (3,) The C function.
        mu: (optional) The gravitational parameter.

    Returns:
        The State Transition Matrix for position with respect to initial velocity.

    Notes:
        The State Transition Matrix for position with respect to initial velocity is defined as:
        $$
        \frac{d\boldsymbol{r}}{d\boldsymbol{v}_0} =
        \frac{r_0}{\mu} (1 - F) [(\boldsymbol{r} - \boldsymbol{r}_0) \boldsymbol{v}_0^T - (\boldsymbol{v} - \boldsymbol{v}_0) \boldsymbol{r}_0^T]
        + \frac{C}{\mu} \boldsymbol{v} \boldsymbol{v}_0^T
        + G \boldsymbol{I}
        $$
        where $\boldsymbol{r}$ is the position vector at current time, $\boldsymbol{v}$ is the velocity vector at the current time, $\boldsymbol{r}_0$ is the position vector at the initial time, $\boldsymbol{v}_0$ is the velocity vector at the initial time, $F$ is the Lagrange F function, $G$ is the Lagrange G function, $C$ is the C function, $\mu$ is the gravitational parameter, and $\boldsymbol{I}$ is the identity matrix.

    References:
        Battin, 1999, pp.467.
    """
    return (
        vector_norm(r0_vec)
        / mu
        * (1 - F)
        * (jnp.outer(r_vec - r0_vec, v0_vec) - jnp.outer(v_vec - v0_vec, r0_vec))
        + C / mu * jnp.outer(v_vec, v0_vec)
        + G * jnp.eye(3)
    )


def pvpr0(
    r_vec: ArrayLike,
    v_vec: ArrayLike,
    r0_vec: ArrayLike,
    v0_vec: ArrayLike,
    Ft: ArrayLike,
    C: ArrayLike,
    mu: ArrayLike = 1,
) -> Array:
    r"""The State Transition Matrix for velocity with respect to initial position.

    Args:
        r_vec: (3,) The position vector at the current time.
        v_vec: (3,) The velocity vector at the current time.
        r0_vec: (3,) The position vector at the initial time.
        v0_vec: (3,) The velocity vector at the initial time.
        Ft: (3,) The Lagrange Ft function.
        C: (3,) The C function.
        mu: (optional) The gravitational parameter.

    Returns:
        The State Transition Matrix for velocity with respect to initial position.

    Notes:
        The State Transition Matrix for velocity with respect to initial position is defined as:
        $$
        \begin{split}
        \frac{d\boldsymbol{v}}{d\boldsymbol{r}_0} =
        & -\frac{1}{r_0^2} (\boldsymbol{v} - \boldsymbol{v}_0) \boldsymbol{r}_0^T
        - \frac{1}{r^2} \boldsymbol{r} (\boldsymbol{v} - \boldsymbol{v}_0)^T  \\
        & + F_t \left(
        \boldsymbol{I} - \frac{\boldsymbol{r} \boldsymbol{r}^T}{r^2}
        + \frac{1}{\mu r} (\boldsymbol{r} \boldsymbol{v}^T - \boldsymbol{v} \boldsymbol{r}^T) \boldsymbol{r}  (\boldsymbol{v} - \boldsymbol{v}_0)^T
        \right)
        - \frac{\mu C}{r^2 r_0^2} \boldsymbol{r} \boldsymbol{r}_0^T
        \end{split}
        $$
        where $\boldsymbol{r}$ is the position vector at current time, $\boldsymbol{v}$ is the velocity vector at the current time, $\boldsymbol{r}_0$ is the position vector at the initial time, $\boldsymbol{v}_0$ is the velocity vector at the initial time, $F_t$ is the Lagrange Ft function, $C$ is the C function, $\mu$ is the gravitational parameter, and $\boldsymbol{I}$ is the identity matrix.

    References:
        Battin, 1999, pp.467.
    """
    r0_mag = vector_norm(r0_vec)
    r_mag = vector_norm(r_vec)
    return (
        -jnp.outer(v_vec - v0_vec, r0_vec / r0_mag**2)
        - jnp.outer(r_vec / r_mag**2, v_vec - v0_vec)
        + Ft
        * (
            jnp.eye(3)
            - jnp.outer(r_vec / r_mag, r_vec / r_mag)
            + jnp.matmul(
                (jnp.outer(r_vec, v_vec) - jnp.outer(v_vec, r_vec)) / mu,
                jnp.outer(r_vec / r_mag, v_vec - v0_vec),
            )
        )
        - mu / r_mag**2 * C / r0_mag**2 * jnp.outer(r_vec / r_mag, r0_vec / r0_mag)
    )


def pvpv0(
    r_vec: ArrayLike,
    v_vec: ArrayLike,
    r0_vec: ArrayLike,
    v0_vec: ArrayLike,
    F: ArrayLike,
    Gt: ArrayLike,
    C: ArrayLike,
    mu: ArrayLike = 1,
) -> Array:
    r"""The State Transition Matrix for velocity with respect to initial velocity.

    Args:
        r_vec: (3,) The position vector at the current time.
        v_vec: (3,) The velocity vector at the current time.
        r0_vec: (3,) The position vector at the initial time.
        v0_vec: (3,) The velocity vector at the initial time.
        F: (3,) The Lagrange F function.
        Gt: (3,) The Lagrange Gt function.
        C: (3,) The C function.
        mu: (optional) The gravitational parameter.

    Returns:
        The State Transition Matrix for velocity with respect to initial velocity.

    Notes:
        The State Transition Matrix for velocity with respect to initial velocity is defined as:
        $$
        \frac{d\boldsymbol{v}}{d\boldsymbol{v}_0} =
        \frac{r_0}{\mu} (\boldsymbol{v} - \boldsymbol{v}_0) (\boldsymbol{v} - \boldsymbol{v}_0)^T
        + \frac{1}{r_0^3}[r_0 (1-F) \boldsymbol{r} \boldsymbol{r}_0^T - C \boldsymbol{r} \boldsymbol{v}_0^T]
        + G_t \boldsymbol{I}
        $$
    """
    r0_mag = vector_norm(r0_vec)
    r_mag = vector_norm(r_vec)
    return (
        r0_mag / mu * jnp.outer(v_vec - v0_vec, v_vec - v0_vec)
        + (1 - F) * jnp.outer(r_vec / r_mag, r0_vec / r_mag) * r0_mag / r_mag
        - C / r_mag**2 * jnp.outer(r_vec / r_mag, v0_vec)
        + Gt * jnp.eye(3)
    )


def dxdx0(
    r_vec: ArrayLike,
    v_vec: ArrayLike,
    r0_vec: ArrayLike,
    v0_vec: ArrayLike,
    deltat: ArrayLike = 3.14,
    mu: ArrayLike = 1,
) -> Array:
    r"""The State Transition Matrix.

    Args:
        r_vec: (3,) The position vector at the current time.
        v_vec: (3,) The velocity vector at the current time.
        r0_vec: (3,) The position vector at the initial time.
        v0_vec: (3,) The velocity vector at the initial time.
        deltat: (optional) The time since the initial time.
        mu: (optional) The gravitational parameter.

    Returns:
        The State Transition Matrix.

    Notes:
        The State Transition Matrix is defined as:
        $$
        \frac{d\boldsymbol{x}}{d\boldsymbol{x}_0} =
        \begin{bmatrix}
        \frac{d\boldsymbol{r}}{d\boldsymbol{r}_0} & \frac{d\boldsymbol{r}}{d\boldsymbol{v}_0} \\
        \frac{d\boldsymbol{v}}{d\boldsymbol{r}_0} & \frac{d\boldsymbol{v}}{d\boldsymbol{v}_0}
        \end{bmatrix}
        $$
        where $\boldsymbol{x} = [\boldsymbol{r}, \boldsymbol{v}]^T$ is the state vector, $\boldsymbol{x}_0 = [\boldsymbol{r}_0, \boldsymbol{v}_0]^T$ is the initial state vector, and $\frac{d\boldsymbol{x}}{d\boldsymbol{x}_0}$ is the State Transition Matrix.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> r_vec = jnp.array([-0.24986234273434585, -0.69332384278075210, 4.9599012168662551e-3])
        >>> v_vec = jnp.array([ 1.2189179487500401,  0.05977450696618754, -0.007101943980682161])
        >>> r0_vec = jnp.array([-0.66234662571997105, 0.74919751798749190, -1.6259997018919074e-4])
        >>> v0_vec = jnp.array([-0.8166746784630675, -0.32961417380268476,  0.006248107587795581])
        >>> deltat = 2.5803148345055149
        >>> mu = 1.0
        >>> expected = jnp.array([[ 6.96499107e+00, -4.49913836e+00, -1.89246497e-02,  7.59769798e+00, -5.23072188e-01, -3.53190183e-02,],
        ...                       [ 1.51857748e+00, -1.52804032e+00, -3.60174714e-03,  2.52883087e+00,  1.17277875e+00, -1.61237785e-02,],
        ...                       [-4.98856694e-02,  2.99417750e-02, -5.82723061e-01, -5.10209945e-02,  8.87921668e-04,  7.78936443e-01,],
        ...                       [ 4.75150081e+00, -3.53238392e+00, -1.06813429e-02,  6.32244923e+00,  1.13091953e+00, -4.74582385e-02,],
        ...                       [ 8.76182663e+00, -6.69223965e+00, -1.65955664e-02,  1.00231736e+01, -5.74210092e-02, -6.14800215e-02,],
        ...                       [-7.39347940e-02,  5.19339444e-02, -4.25016825e-01, -9.32086959e-02, -1.19134537e-02, -1.14713870e+00]])
        >>> jnp.allclose(adx.twobody.dxdx0(r_vec, v_vec, r0_vec, v0_vec, deltat), expected)
        Array(True, dtype=bool)
    """
    r0_mag, r_mag = vector_norm(r0_vec), vector_norm(r_vec)
    alpha = 1.0 / adx.semimajor_axis(r0_mag, vector_norm(v0_vec), mu)
    chi = adx.generalized_anomaly(
        alpha, sigma_bvp(r_vec, v_vec, mu), sigma_bvp(r0_vec, v0_vec, mu), deltat, mu
    )
    C = _C_func(
        chi, ufunc2(chi, alpha), ufunc4(chi, alpha), ufunc5(chi, alpha), deltat, mu
    )
    F = lagrange_F(ufunc2(chi, alpha), r0_mag)
    G = lagrange_G(
        ufunc1(chi, alpha),
        ufunc2(chi, alpha),
        sigma_bvp(r0_vec, v0_vec, mu),
        r0_mag,
        mu,
    )
    Ft = lagrange_Ft(ufunc1(chi, alpha), r_mag, r0_mag, mu)
    Gt = lagrange_Gt(ufunc2(chi, alpha), r_mag)
    return jnp.block(
        [
            [
                prpr0(r_vec, v_vec, r0_vec, v0_vec, F, C, mu),
                prpv0(r_vec, v_vec, r0_vec, v0_vec, F, G, C, mu),
            ],
            [
                pvpr0(r_vec, v_vec, r0_vec, v0_vec, Ft, C, mu),
                pvpv0(r_vec, v_vec, r0_vec, v0_vec, F, Gt, C, mu),
            ],
        ]
    )



from astrodynx.twobody._lagrange import (
    lagrange_F,
    lagrange_G,
    lagrange_Ft,
    lagrange_Gt,
)
from astrodynx.twobody._uniformulas import (
    sigma_bvp,
    ufunc0,
    ufunc1,
    ufunc2,
    ufunc3,
    ufunc4,
    ufunc5,
)
from astrodynx.twobody._state_trans import (
    prpr0,
    prpv0,
    pvpr0,
    pvpv0,
    dxdx0,
)
from astrodynx.twobody._path_check import (
    nmax_by_periapsis,
    is_short_way,
    pass_perigee,
    rp_islower_rmin,
)

__all__ = [
    "lagrange_F",
    "lagrange_G",
    "lagrange_Ft",
    "lagrange_Gt",
    "prpr0",
    "prpv0",
    "pvpr0",
    "pvpv0",
    "dxdx0",
    "sigma_bvp",
    "ufunc0",
    "ufunc1",
    "ufunc2",
    "ufunc3",
    "ufunc4",
    "ufunc5",
    "nmax_by_periapsis",
    "is_short_way",
    "pass_perigee",
    "rp_islower_rmin",
]



import astrodynx as adx
import jax
import jax.numpy as jnp


class TestKeplerEquElps:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        E = jnp.pi / 4
        e = 0.1
        M = 0.7
        expected = E - e * jnp.sin(E) - M
        result = adx.kepler_equ_elps(E, e, M)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        E = jnp.array([jnp.pi / 4, jnp.pi / 2])
        e = jnp.array([0.1, 0.2])
        M = jnp.array([0.7, 0.8])
        expected = E - e * jnp.sin(E) - M
        result = adx.kepler_equ_elps(E, e, M)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        E = jnp.array([jnp.pi / 4, jnp.pi / 2])
        e = 0.1
        M = 0.7
        expected = E - e * jnp.sin(E) - M
        result = adx.kepler_equ_elps(E, e, M)
        assert jnp.allclose(result, expected)

    def test_circular_orbit(self) -> None:
        """Test with circular orbit (e=0)."""
        E = jnp.linspace(0, 2 * jnp.pi, 10)
        e = 0.0
        M = E  # For circular orbits, E = M
        expected = jnp.zeros_like(E)
        result = adx.kepler_equ_elps(E, e, M)
        assert jnp.allclose(result, expected)


class TestKeplerEquHypb:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        H = 1.0
        e = 1.5
        N = 1.0
        expected = e * jnp.sinh(H) - H - N
        result = adx.kepler_equ_hypb(H, e, N)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        H = jnp.array([1.0, 2.0])
        e = jnp.array([1.5, 1.5])
        N = jnp.array([1.0, 1.0])
        expected = e * jnp.sinh(H) - H - N
        result = adx.kepler_equ_hypb(H, e, N)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        H = jnp.array([1.0, 2.0])
        e = 1.5
        N = 1.0
        expected = e * jnp.sinh(H) - H - N
        result = adx.kepler_equ_hypb(H, e, N)
        assert jnp.allclose(result, expected)

    def test_different_eccentricities(self) -> None:
        """Test with different eccentricity values."""
        H = 1.0
        e_values = jnp.array([1.1, 1.5, 2.0])
        N = 1.0
        expected = e_values * jnp.sinh(H) - H - N
        results = jnp.stack([adx.kepler_equ_hypb(H, e, N) for e in e_values])
        assert jnp.allclose(results, expected)


class TestMeanAnomalyElps:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        a = 1.0
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / a**3) * deltat
        result = adx.mean_anomaly_elps(a, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        a = jnp.array([1.0, 2.0])
        mu = jnp.array([1.0, 2.0])
        deltat = jnp.array([1.0, 1.0])
        expected = jnp.sqrt(mu / a**3) * deltat
        result = adx.mean_anomaly_elps(a, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        a = jnp.array([1.0, 2.0])
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / a**3) * deltat
        result = adx.mean_anomaly_elps(a, deltat, mu)
        assert jnp.allclose(result, expected)


class TestMeanAnomalyHypb:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        a = -1.0
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / -(a**3)) * deltat
        result = adx.mean_anomaly_hypb(a, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        a = jnp.array([-1.0, -2.0])
        mu = jnp.array([1.0, 2.0])
        deltat = jnp.array([1.0, 1.0])
        expected = jnp.sqrt(mu / -(a**3)) * deltat
        result = adx.mean_anomaly_hypb(a, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        a = jnp.array([-1.0, -2.0])
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / -(a**3)) * deltat
        result = adx.mean_anomaly_hypb(a, deltat, mu)
        assert jnp.allclose(result, expected)


class TestKeplerEquUni:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        chi = 1.0
        alpha = 1.0
        r0 = 1.0
        sigma0 = 0.0
        deltat = 1.0
        mu = 1.0
        expected = (
            r0 * adx.twobody.ufunc1(chi, alpha)
            + sigma0 * adx.twobody.ufunc2(chi, alpha)
            + adx.twobody.ufunc3(chi, alpha)
            - jnp.sqrt(mu) * deltat
        )
        result = adx.kepler_equ_uni(chi, alpha, r0, sigma0, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        chi = jnp.array([1.0, 2.0])
        alpha = 1.0
        r0 = jnp.array([1.0, 1.0])
        sigma0 = jnp.array([0.0, 0.0])
        deltat = jnp.array([1.0, 1.0])
        mu = jnp.array([1.0, 1.0])
        expected = (
            r0 * adx.twobody.ufunc1(chi, alpha)
            + sigma0 * adx.twobody.ufunc2(chi, alpha)
            + adx.twobody.ufunc3(chi, alpha)
            - jnp.sqrt(mu) * deltat
        )
        result = adx.kepler_equ_uni(chi, alpha, r0, sigma0, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_gradient_wrt_chi(self) -> None:
        """Test gradient with respect to chi."""
        chi = jnp.pi * 0.6
        alpha = jnp.array([1.0, 0, -1.0])
        r0, sigma0, deltat, mu = 1.0, 0.0, 1.0, 1.0
        U0 = jax.vmap(adx.twobody.ufunc0, in_axes=(None, 0))(chi, alpha)
        U1 = jax.vmap(adx.twobody.ufunc1, in_axes=(None, 0))(chi, alpha)
        U2 = jax.vmap(adx.twobody.ufunc2, in_axes=(None, 0))(chi, alpha)
        expected = adx.equ_of_orb_uvi(U0, U1, U2, r0, sigma0) / jnp.sqrt(
            mu
        )  # dt/d chi = r / sqrt(mu), Battin 1999, pp.174.

        result = jax.vmap(
            jax.grad(adx.kepler_equ_uni, argnums=0),
            in_axes=(None, 0, None, None, None, None),
        )(chi, alpha, r0, sigma0, deltat, mu)
        assert jnp.allclose(result, expected)


class TestGeneralizedAnomaly:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        alpha = 1.0
        sigma = 1.0
        sigma0 = 0.0
        deltat = 1.0
        mu = 1.0
        expected = alpha * jnp.sqrt(mu) * deltat + sigma - sigma0
        result = adx.generalized_anomaly(alpha, sigma, sigma0, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        alpha = jnp.array([1.0, 1.0])
        sigma = jnp.array([1.0, 2.0])
        sigma0 = jnp.array([0.0, 0.0])
        deltat = jnp.array([1.0, 1.0])
        mu = jnp.array([1.0, 1.0])
        expected = alpha * jnp.sqrt(mu) * deltat + sigma - sigma0
        result = adx.generalized_anomaly(alpha, sigma, sigma0, deltat, mu)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        alpha = 1.0
        sigma = jnp.array([1.0, 2.0])
        sigma0 = 0.0
        deltat = 1.0
        mu = 1.0
        expected = alpha * jnp.sqrt(mu) * deltat + sigma - sigma0
        result = adx.generalized_anomaly(alpha, sigma, sigma0, deltat, mu)
        assert jnp.allclose(result, expected)



import jax.numpy as jnp
from astrodynx.twobody import pass_perigee


class TestPassPerigee:
    r1 = jnp.array([8081.226263, -0.0, -0.0])
    v1 = jnp.array([-0.440773, 7.6, 0.133593])
    mu = 3.986e5

    def test_pass_perigee1(self):
        r2 = jnp.array([7902.679945, 1371.949935, 23.947475])
        assert not pass_perigee(self.r1, self.v1, r2, self.mu)
        assert pass_perigee(self.r1, -self.v1, r2, self.mu)

    def test_pass_perigee2(self):
        r2 = jnp.array([-10664.744693, 3855.689988, 67.301319])
        assert pass_perigee(self.r1, self.v1, r2, self.mu)
        assert not pass_perigee(self.r1, -self.v1, r2, self.mu)

    def test_pass_perigee3(self):
        r2 = jnp.array([-3746.681531, -10353.987124, -180.729518])
        assert pass_perigee(self.r1, self.v1, r2, self.mu)
        assert not pass_perigee(self.r1, -self.v1, r2, self.mu)

    def test_pass_perigee4(self):
        r2 = jnp.array([6742.310614, -5513.347909, -96.235846])
        assert pass_perigee(self.r1, self.v1, r2, self.mu)
        assert not pass_perigee(self.r1, -self.v1, r2, self.mu)

    def test_basic_case(self) -> None:
        """Test with a basic circular orbit case."""
        r1 = jnp.array([1.0, 0.0, 0.0])
        v1 = jnp.array([0.0, 1.0, 0.0])
        r2 = jnp.array([0.0, 1.0, 0.0])
        assert not pass_perigee(r1, v1, r2)

    def test_elliptical_orbit(self) -> None:
        """Test with an elliptical orbit where perigee is passed."""
        # Create an elliptical orbit with e=0.5
        r1 = jnp.array([1.5, 0.0, 0.0])  # Starting at apogee
        v1 = jnp.array([0.0, 0.8, 0.0])  # Velocity for elliptical orbit
        r2 = jnp.array([-1.0, 0.0, 0.0])  # Position after passing perigee
        assert pass_perigee(r1, v1, r2)



import jax.numpy as jnp
from jax.typing import ArrayLike
from jax import Array

"""Orbital integrals and elements for two-body orbital mechanics."""


def orb_period(a: ArrayLike, mu: ArrayLike = 1) -> Array:
    r"""
    Returns the orbital period of a two-body system.

    Args:
        a: Semimajor axis of the orbit.
        mu: Gravitational parameter of the central body; shape broadcast-compatible with `a`.

    Returns:
        The orbital period of the object in the two-body system.

    Notes:
        The orbital period is calculated using Kepler's third law:
        $$
        P = 2\pi \sqrt{\frac{a^3}{\mu}}
        $$
        where $P$ is the orbital period, $a$ is the semimajor axis, and $\mu$ is the gravitational parameter.

    References:
        Battin, 1999, pp.119.

    Examples:
        A simple example of calculating the orbital period for a circular orbit with a semimajor axis of 1.0 and a gravitational parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> a = 1.0
        >>> mu = 1.0
        >>> adx.orb_period(a, mu)
        Array(6.2831855, dtype=float32, weak_type=True)

        With broadcasting, you can calculate the orbital period for multiple semimajor axes and gravitational parameters:

        >>> a = jnp.array([1.0, 2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> adx.orb_period(a, mu)
        Array([ 6.2831855, 12.566371 ], dtype=float32)
    """
    return 2 * jnp.pi * jnp.sqrt(a**3 / mu)


def angular_momentum(pos_vec: ArrayLike, vel_vec: ArrayLike) -> Array:
    r"""
    Returns the specific angular momentum of a two-body system.

    Args:
        pos_vec: (..., 3) position vector of the object in the two-body system.
        vel_vec: (..., 3) velocity vector of the object in the two-body system, which shape broadcast-compatible with `pos_vec`.

    Returns:
        The specific angular momentum vector of the object in the two-body system.

    Notes
        The specific angular momentum is calculated using the cross product of the position and velocity vectors:
        $$
        \boldsymbol{h} = \boldsymbol{r} \times \boldsymbol{v}
        $$
        where $\boldsymbol{h}$ is the specific angular momentum, $\boldsymbol{r}$ is the position vector, and $\boldsymbol{v}$ is the velocity vector.

    References
        Battin, 1999, pp.115.

    Examples
        A simple example of calculating the specific angular momentum for a position vector [1, 0, 0] and velocity vector [0, 1, 0]:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> pos_vec = jnp.array([1.0, 0.0, 0.0])
        >>> vel_vec = jnp.array([0.0, 1.0, 0.0])
        >>> adx.angular_momentum(pos_vec, vel_vec)
        Array([0., 0., 1.], dtype=float32)

        With broadcasting, you can calculate the specific angular momentum for multiple position and velocity vectors:

        >>> pos_vec = jnp.array([[1.0, 0.0, 0.0], [2.0, 0.0, 0.0]])
        >>> vel_vec = jnp.array([[0.0, 1.0, 0.0], [0.0, 2.0, 0.0]])
        >>> adx.angular_momentum(pos_vec, vel_vec)
        Array([[0., 0., 1.],
               [0., 0., 4.]], dtype=float32)
    """
    return jnp.cross(pos_vec, vel_vec)


def semimajor_axis(r_mag: ArrayLike, v_mag: ArrayLike, mu: ArrayLike = 1) -> ArrayLike:
    r"""
    Returns the semimajor axis of a two-body orbit.

    Args:
        r_mag: Norm of the object's position vector in the two-body system.
        v_mag: Norm of the object's velocity vector in the two-body system, which shape broadcast-compatible with `r`.
        mu: Gravitational parameter of the central body; shape broadcast-compatible with `r` and `v`.

    Returns:
        The semimajor axis of the orbit.

    Notes
        The semimajor axis is calculated using equation (3.16):
        $$
        a = \left( \frac{2}{r} - \frac{v^2}{\mu} \right)^{-1}
        $$
        where $a$ is the semimajor axis, $r$ is the norm of the position vector, $v$ is the norm of the velocity vector, and $\mu$ is the gravitational parameter.

    References
        Battin, 1999, pp.116.

    Examples
        A simple example of calculating the semimajor axis with a position vector norm of 1.0, velocity vector norm of 1.0, and gravitational parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> r = 1.0
        >>> v = 1.0
        >>> mu = 1.0
        >>> adx.semimajor_axis(r, v, mu)
        1.0

        With broadcasting, you can calculate the semimajor axis for multiple position and velocity vectors:

        >>> r = jnp.array([1.0, 2.0])
        >>> v = jnp.array([1.0, 2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> adx.semimajor_axis(r, v, mu)
        Array([ 1., -1.], dtype=float32)
    """
    return 1 / (2 / r_mag - v_mag**2 / mu)


def eccentricity_vector(
    pos_vec: ArrayLike, vel_vec: ArrayLike, mu: ArrayLike = 1
) -> Array:
    r"""
    Returns the eccentricity vector of a two-body orbit.

    Args:
        pos_vec: (..., 3) position vector of the object in the two-body system.
        vel_vec: (..., 3) velocity vector of the object in the two-body system, which shape broadcast-compatible with `pos_vec`.
        mu: Gravitational parameter of the central body; shape broadcast-compatible with `pos_vec` and `vel_vec`.

    Returns:
        The eccentricity vector of the orbit.

    Notes
        The eccentricity vector is calculated using equation (3.14):
        $$
        \boldsymbol{e} = \frac{\boldsymbol{v} \times \boldsymbol{h}}{\mu} - \frac{\boldsymbol{r}}{r}
        $$
        where $\boldsymbol{e}$ is the eccentricity vector, $\boldsymbol{v}$ is the velocity vector, $\boldsymbol{h}$ is the specific angular momentum vector, $\mu$ is the gravitational parameter, and $\boldsymbol{r}$ is the position vector.

    References
        Battin, 1999, pp.116.

    Examples
        A simple example of calculating the eccentricity vector for a position vector [1, 0, 0] and velocity vector [0, 1, 0]:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> pos_vec = jnp.array([1.0, 0.0, 0.0])
        >>> vel_vec = jnp.array([0.0, 1.0, 0.0])
        >>> mu = 1.0
        >>> adx.eccentricity_vector(pos_vec, vel_vec, mu)
        Array([0., 0., 0.], dtype=float32)

        With broadcasting, you can calculate the eccentricity vector for multiple position and velocity vectors:

        >>> pos_vec = jnp.array([[1.0, 0.0, 0.0], [2.0, 0.0, 0.0]])
        >>> vel_vec = jnp.array([[0.0, 1.0, 0.0], [0.0, 2.0, 0.0]])
        >>> mu = jnp.array([[1.0],[2.0]])
        >>> adx.eccentricity_vector(pos_vec, vel_vec, mu)
        Array([[0., 0., 0.],
               [3., 0., 0.]], dtype=float32)
    """
    h = angular_momentum(pos_vec, vel_vec)
    return jnp.cross(vel_vec, h) / mu - pos_vec / jnp.linalg.vector_norm(
        pos_vec, axis=-1, keepdims=True
    )


def semiparameter(h_mag: ArrayLike, mu: ArrayLike = 1) -> ArrayLike:
    r"""
    Returns the semiparameter of a two-body orbit.

    Args:
        h_mag: The angular momentum of the object in the two-body system.
        mu: Gravitational parameter of the central body; shape broadcast-compatible with `h`.

    Returns:
        The semiparameter of the orbit.

    Notes
        The semiparameter is calculated using equation (3.15):
        $$
        p = \frac{h^2}{\mu}
        $$
        where $p$ is the semiparameter, $h$ is the norm of the specific angular momentum vector, and $\mu$ is the gravitational parameter.

    References
        Battin, 1999, pp.116.

    Examples
        A simple example of calculating the semiparameter with a specific angular momentum norm of 1.0 and gravitational parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> h = 1.0
        >>> mu = 1.0
        >>> adx.semiparameter(h, mu)
        1.0

        With broadcasting, you can calculate the semiparameter for multiple specific angular momentum norms and gravitational parameters:

        >>> h = jnp.array([1.0, 2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> adx.semiparameter(h, mu)
        Array([1., 2.], dtype=float32)
    """
    return h_mag**2 / mu


def mean_motion(P: ArrayLike) -> ArrayLike:
    r"""
    Returns the mean motion of a two-body orbit.

    Args:
        P: Orbital period of the object in the two-body system.

    Returns:
        The mean motion of the orbit.

    Notes
        The mean motion is calculated using equation (3.24):
        $$
        n = \frac{2\pi}{P}
        $$
        where $n$ is the mean motion and $P$ is the orbital period.

    References
        Battin, 1999, pp.119.

    Examples
        A simple example of calculating the mean motion with an orbital period of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> P = 1.0
        >>> adx.mean_motion(P)
        6.2831...

        With broadcasting, you can calculate the mean motion for multiple orbital periods:

        >>> P = jnp.array([1.0, 2.0])
        >>> adx.mean_motion(P)
        Array([6.2831..., 3.1415...], dtype=float32)
    """
    return 2 * jnp.pi / P


def equ_of_orbit(p: ArrayLike, e: ArrayLike, f: ArrayLike) -> Array:
    r"""
    Returns the equation of the orbit in the two-body system.

    Args:
        p: Semiparameter of the orbit.
        e: Eccentricity of the orbit; shape broadcast-compatible with `p`.
        f: True anomaly of the orbit; shape broadcast-compatible with `p` and `e`.

    Returns:
        The equation of the orbit.

    Notes
        The equation of the orbit is calculated using equation (3.20):
        $$
        r = \frac{p}{1 + e \cos(f)}
        $$
        where $r$ is the norm of the position vector, $p$ is the semiparameter, $e$ is the eccentricity, and $f$ is the true anomaly.

    References
        Battin, 1999, pp.117.

    Examples
        A simple example of calculating the equation of the orbit with a semiparameter of 1.0, eccentricity of 0.0, and true anomaly of 0.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> p = 1.0
        >>> e = 0.0
        >>> f = 0.0
        >>> adx.equ_of_orbit(p, e, f)
        Array(1., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the equation of the orbit for multiple semiparameters, eccentricities, and true anomalies:

        >>> p = jnp.array([1.0, 2.0])
        >>> e = jnp.array([0.0, 0.0])
        >>> f = jnp.array([0.0, 0.0])
        >>> adx.equ_of_orbit(p, e, f)
        Array([1., 2.], dtype=float32)
    """
    return p / (1 + e * jnp.cos(f))


def equ_of_orb_uvi(
    U0: ArrayLike, U1: ArrayLike, U2: ArrayLike, r0: ArrayLike, sigma0: ArrayLike
) -> ArrayLike:
    r"""The equation of orbit using universal functions

    Args:
        U0: The universal function U0.
        U1: The universal function U1.
        U2: The universal function U2.
        r0: The radius at the initial time.
        sigma0: The sigma function at the initial time.

    Returns:
        The radius.

    Notes:
        The function is defined as:
        $$
        r = r_0 U_0 + \sigma_0 U_1 + U_2
        $$
        where $U_0$ is the universal function U0, $U_1$ is the universal function U1, $U_2$ is the universal function U2, $r_0$ is the radius at the initial time, and $\sigma_0$ is the sigma function at the initial time.

    References:
        Battin, 1999, pp.178.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> U0 = 1.0
        >>> U1 = 1.0
        >>> U2 = 1.0
        >>> r0 = 1.0
        >>> sigma0 = 0.0
        >>> adx.equ_of_orb_uvi(U0, U1, U2, r0, sigma0)
        2.0

        With broadcasting:

        >>> U0 = jnp.array([1.0, 2.0])
        >>> U1 = jnp.array([1.0, 2.0])
        >>> U2 = jnp.array([1.0, 1.0])
        >>> r0 = jnp.array([1.0, 1.0])
        >>> sigma0 = jnp.array([0.0, 0.0])
        >>> adx.equ_of_orb_uvi(U0, U1, U2, r0, sigma0)
        Array([2., 3.], dtype=float32)
    """
    return r0 * U0 + sigma0 * U1 + U2


def radius_periapsis(p: ArrayLike, e: ArrayLike) -> ArrayLike:
    r"""
    Returns the radius of periapsis of the orbit.

    Args:
        p: Semiparameter of the orbit.
        e: Eccentricity of the orbit; shape broadcast-compatible with `p`.

    Returns:
        The radius of periapsis of the orbit.

    Notes
        The radius of periapsis is calculated using equation:
        $$
        r_p = \frac{p}{1 + e}
        $$
        where $r_p$ is the radius of periapsis, $p$ is the semiparameter, and $e$ is the eccentricity.

    References
        Battin, 1999, pp.117.

    Examples
        A simple example of calculating the radius of periapsis with a semiparameter of 1.0 and eccentricity of 0.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> p = 1.0
        >>> e = 0.0
        >>> adx.radius_periapsis(p, e)
        1.0

        With broadcasting, you can calculate the radius of periapsis for multiple semiparameters and eccentricities:

        >>> p = jnp.array([1.0, 2.0])
        >>> e = jnp.array([0.0, 0.0])
        >>> adx.radius_periapsis(p, e)
        Array([1., 2.], dtype=float32)
    """
    return p / (1 + e)


def radius_apoapsis(p: ArrayLike, e: ArrayLike) -> ArrayLike:
    r"""
    Returns the radius of apoapsis of the orbit.

    Args:
        p: Semiparameter of the orbit.
        e: Eccentricity of the orbit; shape broadcast-compatible with `p`.

    Returns:
        The radius of apoapsis of the orbit.

    Notes
        The radius of apoapsis is calculated using equation:
        $$
        r_a = \frac{p}{1 - e}
        $$
        where $r_a$ is the radius of apoapsis, $p$ is the semiparameter, and $e$ is the eccentricity.

    References
        Battin, 1999, pp.117.

    Examples
        A simple example of calculating the radius of apoapsis with a semiparameter of 1.0 and eccentricity of 0.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> p = 1.0
        >>> e = 0.0
        >>> adx.radius_apoapsis(p, e)
        1.0

        With broadcasting, you can calculate the radius of apoapsis for multiple semiparameters and eccentricities:

        >>> p = jnp.array([1.0, 2.0])
        >>> e = jnp.array([0.0, 0.0])
        >>> adx.radius_apoapsis(p, e)
        Array([1., 2.], dtype=float32)
    """
    return p / (1 - e)



import jax.numpy as jnp
from jax.typing import ArrayLike, DTypeLike
from jax import Array
from astrodynx.twobody._uniformulas import ufunc1, ufunc2, ufunc3

"""Kepler's equations and generalized anomaly for two-body orbital mechanics."""


def kepler_equ_elps(E: ArrayLike, e: ArrayLike, M: ArrayLike = 0) -> Array:
    r"""Returns the Kepler's equation for elliptical orbits in the form f(E) = 0.

    Args:
        E: Eccentric anomaly.
        e: Eccentricity of the orbit, 0 <= e < 1.
        M: (optional) Mean anomaly.

    Returns:
        The value of Kepler's equation for elliptical orbits: E - e*sin(E) - M.

    Notes:
        Kepler's equation for elliptical orbits relates the eccentric anomaly E to the mean anomaly M:
        $$
        E - e \sin E = M
        $$
        This function returns the equation in the form f(E) = 0, which is useful for root-finding algorithms.

    References:
        Battin, 1999, pp.160.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> E = jnp.pi/4
        >>> e = 0.1
        >>> M = 0.7
        >>> adx.kepler_equ_elps(E, e, M)
        Array(0.01468..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the Kepler's equation for multiple eccentric anomalies, eccentricities, and mean anomalies:

        >>> E = jnp.array([jnp.pi/4, jnp.pi/2])
        >>> e = jnp.array([0.1, 0.2])
        >>> M = jnp.array([0.7, 0.8])
        >>> adx.kepler_equ_elps(E, e, M)
        Array([0.01468..., 0.5707...], dtype=float32)
    """
    return E - e * jnp.sin(E) - M


def kepler_equ_hypb(H: ArrayLike, e: ArrayLike, N: ArrayLike = 0) -> Array:
    r"""Returns the Kepler's equation for hyperbolic orbits in the form f(H) = 0.

    Args:
        H: Hyperbolic eccentric anomaly.
        e: Eccentricity of the orbit, e > 1.
        N: (optional) Hyperbolic mean anomaly.

    Returns:
        The value of Kepler's equation for hyperbolic orbits: e*sinh(H) - H - N.

    Notes:
        Kepler's equation for hyperbolic orbits relates the hyperbolic eccentric anomaly H to the hyperbolic mean anomaly N:
        $$
        e \sinh H - H = N
        $$
        This function returns the equation in the form f(H) = 0, which is useful for root-finding algorithms.

    References:
        Battin, 1999, pp.168.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> H = 1.0
        >>> e = 1.5
        >>> N = 1.0
        >>> adx.kepler_equ_hypb(H, e, N)
        Array(-0.2371..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the Kepler's equation for multiple hyperbolic eccentric anomalies, eccentricities, and hyperbolic mean anomalies:

        >>> H = jnp.array([1.0, 2.0])
        >>> e = jnp.array([1.5, 1.5])
        >>> N = jnp.array([1.0, 1.0])
        >>> adx.kepler_equ_hypb(H, e, N)
        Array([-0.2371...,  2.4402...], dtype=float32)
    """
    return e * jnp.sinh(H) - H - N


def mean_anomaly_elps(a: ArrayLike, deltat: ArrayLike, mu: ArrayLike = 1) -> Array:
    r"""Returns the mean anomaly for an elliptical orbit.

    Args:
        a: Semimajor axis of the orbit, a > 0.
        deltat: Time since periapsis passage.
        mu: (optional) Gravitational parameter of the central body.

    Returns:
        The mean anomaly for an elliptical orbit.

    Notes:
        The mean anomaly for an elliptical orbit is calculated using the formula:
        $$
        M = \sqrt{\frac{\mu}{a^3}} \Delta t
        $$
        where $M$ is the mean anomaly, $a>0$ is the semimajor axis, $\mu$ is the gravitational parameter, and $\Delta t$ is the time since periapsis passage.

    References:
        Battin, 1999, pp.160.

    Examples:
        A simple example of calculating the mean anomaly for an orbit with semimajor axis 1.0, gravitational parameter 1.0, and time since periapsis passage 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> a = 1.0
        >>> mu = 1.0
        >>> deltat = 1.0
        >>> adx.mean_anomaly_elps(a, deltat, mu)
        Array(1., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the mean anomaly for multiple semimajor axes, gravitational parameters, and times since periapsis passage:

        >>> a = jnp.array([1.0, 2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> deltat = jnp.array([1.0, 1.0])
        >>> adx.mean_anomaly_elps(a, deltat, mu)
        Array([1. , 0.5], dtype=float32)
    """
    return jnp.sqrt(mu / a**3) * deltat


def mean_anomaly_hypb(a: ArrayLike, deltat: ArrayLike, mu: ArrayLike = 1) -> Array:
    r"""Returns the mean anomaly for a hyperbolic orbit.

    Args:
        a: Semimajor axis of the orbit, a < 0.
        deltat: Time since periapsis passage.
        mu: (optional) Gravitational parameter of the central body.

    Returns:
        The mean anomaly for a hyperbolic orbit.

    Notes:
        The mean anomaly for a hyperbolic orbit is calculated using the formula:
        $$
        N = \sqrt{\frac{\mu}{-a^3}} \Delta t
        $$
        where $N$ is the mean anomaly, $a<0$ is the semimajor axis, $\mu$ is the gravitational parameter, and $\Delta t$ is the time since periapsis passage.

    References:
        Battin, 1999, pp.166.

    Examples:
        A simple example of calculating the mean anomaly for an orbit with semimajor axis -1.0, gravitational parameter 1.0, and time since periapsis passage 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> a = -1.0
        >>> mu = 1.0
        >>> deltat = 1.0
        >>> adx.mean_anomaly_hypb(a, deltat, mu)
        Array(1., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the mean anomaly for multiple semimajor axes, gravitational parameters, and times since periapsis passage:

        >>> a = jnp.array([-1.0, -2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> deltat = jnp.array([1.0, 1.0])
        >>> adx.mean_anomaly_hypb(a, deltat, mu)
        Array([1. , 0.5], dtype=float32)
    """
    return jnp.sqrt(mu / -(a**3)) * deltat


def kepler_equ_uni(
    chi: ArrayLike,
    alpha: DTypeLike = 1,
    r0: ArrayLike = 1,
    sigma0: ArrayLike = 0,
    deltat: ArrayLike = 0,
    mu: ArrayLike = 1,
) -> Array:
    r"""Returns the universal Kepler's equation in the form f(chi) = 0.

    Args:
        chi: The generalized anomaly.
        alpha: (optional) The reciprocal of the semimajor axis.
        r0: (optional) The radius at the initial time.
        sigma0: (optional) The sigma function at the initial time.
        deltat: (optional) The time since the initial time.
        mu: (optional) The gravitational parameter.

    Returns:
        The value of the universal Kepler's equation.

    Notes:
        The universal Kepler's equation is defined as:
        $$
        r_0 U_1(\chi, \alpha) + \sigma_0 U_2(\chi, \alpha) + U_3(\chi, \alpha) - \sqrt{\mu} \Delta t = 0
        $$
        where $\chi$ is the generalized anomaly, $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis, $\sigma_0$ is the sigma function at the initial time, $r_0$ is the norm of the position vector at the initial time, $\mu$ is the gravitational parameter, $U_1$ is the universal function U1, $U_2$ is the universal function U2, $U_3$ is the universal function U3, and $\Delta t$ is the time since the initial time.

    References:
        Battin, 1999, pp.178.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> sigma0 = 0.0
        >>> r0 = 1.0
        >>> mu = 1.0
        >>> deltat = 1.0
        >>> adx.kepler_equ_uni(chi, alpha, r0, sigma0, deltat, mu)
        Array(0., dtype=float32, weak_type=True)

        With broadcasting:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = 1.
        >>> sigma0 = jnp.array([0.0, 0.0])
        >>> r0 = jnp.array([1.0, 1.0])
        >>> deltat = jnp.array([1.0, 1.0])
        >>> mu = jnp.array([1.0, 1.0])
        >>> adx.kepler_equ_uni(chi, alpha, r0, sigma0, deltat, mu)
        Array([0., 1.], dtype=float32)
    """
    return (
        r0 * ufunc1(chi, alpha)
        + sigma0 * ufunc2(chi, alpha)
        + ufunc3(chi, alpha)
        - jnp.sqrt(mu) * deltat
    )


def generalized_anomaly(
    alpha: ArrayLike,
    sigma: ArrayLike,
    sigma0: ArrayLike,
    deltat: ArrayLike = 0,
    mu: ArrayLike = 1,
) -> Array:
    r"""Returns the generalized anomaly.

    Args:
        alpha: The reciprocal of the semimajor axis.
        sigma: The sigma function at the current time.
        sigma0: The sigma function at the initial time.
        deltat: (optional) The time since the initial time.
        mu: (optional) The gravitational parameter.

    Returns:
        The generalized anomaly.

    Notes:
        The generalized anomaly is defined as:
        $$
        \chi = \alpha \sqrt{\mu} \Delta t + \sigma - \sigma_0
        $$
        where $\chi$ is the generalized anomaly, $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis, $\sigma$ is the sigma function at the current time, $\sigma_0$ is the sigma function at the initial time, $\mu$ is the gravitational parameter, and $\Delta t$ is the time since the initial time.

    References:
        Battin, 1999, pp.179.

    Examples:
        A simple example:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> alpha = 1.0
        >>> sigma = 1.0
        >>> sigma0 = 0.0
        >>> mu = 1.0
        >>> deltat = 1.0
        >>> adx.generalized_anomaly(alpha, sigma, sigma0, deltat, mu)
        Array(2., dtype=float32, weak_type=True)

        With broadcasting:

        >>> alpha = jnp.array([1.0, 1.0])
        >>> sigma = jnp.array([1.0, 2.0])
        >>> sigma0 = jnp.array([0.0, 0.0])
        >>> mu = jnp.array([1.0, 1.0])
        >>> deltat = jnp.array([1.0, 1.0])
        >>> adx.generalized_anomaly(alpha, sigma, sigma0, deltat, mu)
        Array([2., 3.], dtype=float32)
    """
    return alpha * jnp.sqrt(mu) * deltat + sigma - sigma0



from astrodynx._version import version as version
from astrodynx._version import version_tuple as version_tuple
from astrodynx._version import __version__ as __version__
from astrodynx._version import __version_tuple__ as __version_tuple__

from astrodynx.twobody._kep_equ import (
    kepler_equ_elps,
    kepler_equ_hypb,
    kepler_equ_uni,
    generalized_anomaly,
    mean_anomaly_elps,
    mean_anomaly_hypb,
)
from astrodynx.twobody._orb_integrals import (
    orb_period,
    angular_momentum,
    semimajor_axis,
    eccentricity_vector,
    semiparameter,
    mean_motion,
    equ_of_orbit,
    equ_of_orb_uvi,
    radius_periapsis,
    radius_apoapsis,
)


__all__ = [
    "kepler_equ_elps",
    "kepler_equ_hypb",
    "kepler_equ_uni",
    "generalized_anomaly",
    "mean_anomaly_elps",
    "mean_anomaly_hypb",
    "orb_period",
    "angular_momentum",
    "semimajor_axis",
    "eccentricity_vector",
    "semiparameter",
    "mean_motion",
    "equ_of_orbit",
    "equ_of_orb_uvi",
    "radius_periapsis",
    "radius_apoapsis",
]



import astrodynx as adx
import jax.numpy as jnp


class TestRotmat3dx:
    def test_zero_angle(self):
        """Test that zero angle returns the identity matrix."""
        angle = 0.0
        expected = jnp.eye(3)
        result = adx.utils.rotmat3dx(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_pi_over_two(self):
        """Test rotation by 90 degrees (π/2 radians)."""
        angle = jnp.pi / 2
        expected = jnp.array([[1.0, 0.0, 0.0], [0.0, 0.0, -1.0], [0.0, 1.0, 0.0]])
        result = adx.utils.rotmat3dx(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_pi(self):
        """Test rotation by 180 degrees (π radians)."""
        angle = jnp.pi
        expected = jnp.array([[1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [0.0, 0.0, -1.0]])
        result = adx.utils.rotmat3dx(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_negative_angle(self):
        """Test rotation by a negative angle (-π/2 radians)."""
        angle = -jnp.pi / 2
        expected = jnp.array([[1.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, -1.0, 0.0]])
        result = adx.utils.rotmat3dx(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_broadcasting(self):
        """Test broadcasting with an array of angles."""
        angles = jnp.array([0.0, jnp.pi / 2])
        results = jnp.stack([adx.utils.rotmat3dx(a) for a in angles])
        expected0 = jnp.eye(3)
        expected1 = jnp.array([[1.0, 0.0, 0.0], [0.0, 0.0, -1.0], [0.0, 1.0, 0.0]])
        assert jnp.allclose(results[0], expected0, atol=1e-7)
        assert jnp.allclose(results[1], expected1, atol=1e-7)


class TestRotmat3dy:
    def test_zero_angle(self):
        """Test that zero angle returns the identity matrix."""
        angle = 0.0
        expected = jnp.eye(3)
        result = adx.utils.rotmat3dy(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_pi_over_two(self):
        """Test rotation by 90 degrees (π/2 radians)."""
        angle = jnp.pi / 2
        expected = jnp.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [-1.0, 0.0, 0.0]])
        result = adx.utils.rotmat3dy(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_pi(self):
        """Test rotation by 180 degrees (π radians)."""
        angle = jnp.pi
        expected = jnp.array([[-1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, -1.0]])
        result = adx.utils.rotmat3dy(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_negative_angle(self):
        """Test rotation by a negative angle (-π/2 radians)."""
        angle = -jnp.pi / 2
        expected = jnp.array([[0.0, 0.0, -1.0], [0.0, 1.0, 0.0], [1.0, 0.0, 0.0]])
        result = adx.utils.rotmat3dy(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_broadcasting(self):
        """Test broadcasting with an array of angles."""
        angles = jnp.array([0.0, jnp.pi / 2])
        results = jnp.stack([adx.utils.rotmat3dy(a) for a in angles])
        expected0 = jnp.eye(3)
        expected1 = jnp.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [-1.0, 0.0, 0.0]])
        assert jnp.allclose(results[0], expected0, atol=1e-7)
        assert jnp.allclose(results[1], expected1, atol=1e-7)


class TestRotmat3dz:
    def test_zero_angle(self):
        """Test that zero angle returns the identity matrix."""
        angle = 0.0
        expected = jnp.eye(3)
        result = adx.utils.rotmat3dz(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_pi_over_two(self):
        """Test rotation by 90 degrees (π/2 radians)."""
        angle = jnp.pi / 2
        expected = jnp.array([[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
        result = adx.utils.rotmat3dz(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_pi(self):
        """Test rotation by 180 degrees (π radians)."""
        angle = jnp.pi
        expected = jnp.array([[-1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [0.0, 0.0, 1.0]])
        result = adx.utils.rotmat3dz(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_negative_angle(self):
        """Test rotation by a negative angle (-π/2 radians)."""
        angle = -jnp.pi / 2
        expected = jnp.array([[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
        result = adx.utils.rotmat3dz(angle)
        assert jnp.allclose(result, expected, atol=1e-7)

    def test_broadcasting(self):
        """Test broadcasting with an array of angles."""
        angles = jnp.array([0.0, jnp.pi / 2])
        results = jnp.stack([adx.utils.rotmat3dz(a) for a in angles])
        expected0 = jnp.eye(3)
        expected1 = jnp.array([[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
        assert jnp.allclose(results[0], expected0, atol=1e-7)
        assert jnp.allclose(results[1], expected1, atol=1e-7)
